/*
** Module   : LocalTempTableHelper.java
** Abstract : Helper class which generates local temp table DDL
**
** Copyright (c) 2004-2008, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 CA  20080424 ADD  @38109  Created initial version. Generates DDL for
**                               local temp tables with local indexes.
** 002 CA  20080430 CHG  @38157  Changes due to original incorrect index
**                               creation for the "dirty" database.
** 003 ECF 20080512 CHG  @38503  Changed getIndexes(). Use new signature for
**                               DMOIndex.databaseIndexes().
** 004 ECF 20080611 CHG  @38696  Modified getComputedCols() to handle new
**                               PersistenceException. DMOIndex now throws an
**                               exception if index metadata cannot be
**                               queried.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import java.util.List;
import java.util.Map;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;
import org.hibernate.mapping.*;

/**
 * Helper class which generates and caches SQL phrases to create and drop
 * temporary tables, and to create indexes in those tables.
 * <p>
 * This class is responsible for local temporary tables support: the created
 * tables and indices will be local to the session which created them.
 */
final class LocalTempTableHelper
extends TempTableHelper
{
   /**
    * Constructor.
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    */
   LocalTempTableHelper(P2JDialect dialect, Class<?> dmoIface)
   {
      super(dialect, dmoIface, DatabaseManager.TEMP_TABLE_DB);
   }

   /**
    * This method will return the needed create table string, which is the 
    * "create local temporary table" string.
    * 
    * @return  The create table string.
    */
   protected String getCreateTableStatement()
   {
      return "create local temporary table ";
   }

   /**
    * This method will return an additional suffix to make the index unique
    * across sessions.
    * 
    * @return The additional suffix.
    */
   protected String getAdditionalIndexSuffix()
   {
      return "__${" + UNIQUE_SUFFIX_KEY + "}";
   }

   /**
    * Populate the list of dialect-specific SQL statements used to create
    * any indexes required by the temp table(s).
    * <p>
    * If there are no indices for the required table, a statement to create an 
    * index for the {@link TemporaryBuffer#MULTIPLEX_FIELD_NAME multiplex 
    * column} will be created.
    * 
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    * @param   iter
    *          The iterator for the collection which contains the needed 
    *          indices. 
    *          
    * @return  List of SQL statements.  Will be empty if no indices are
    *          defined.
    */
   protected List<String> generateSQLCreateIndexList(P2JDialect dialect,
                                                     Class<?> dmoIface, 
                                                     Table primary, 
                                                     List<Table> secondaries,
                                                     Iterator<P2JIndex> iter)
   {
      List<String> stmts = new ArrayList<>();
      
      // Handle indices for primary table first.
      if (!iter.hasNext())
      {
         // If no other indices exist, create an index on the multiplex ID
         // of the primary table.
         String tableName = primary.getName();
         String idxName = "idx_mpid__" + tableName;
         P2JIndex idx = new P2JIndex(tableName, idxName, false);
         idx.addComponent(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
         stmts.add(generateSQLCreateIndex(dialect, idx, primary, false, true));
         
         ((ArrayList<String>) stmts).trimToSize();
      }
      else
      {
         stmts = super.generateSQLCreateIndexList(dialect, dmoIface, primary, secondaries, iter);
      }
      
      return stmts;
   }
   
   /**
    * Populate the list of dialect-specific SQL statements used to drop
    * any indexes required by the temp table(s).
    * <p>
    * If there are no indexes for the required table, a statement to drop the 
    * index associated with the {@link TemporaryBuffer#MULTIPLEX_FIELD_NAME 
    * multiplex column} will be created.
    * 
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    * @param   iter
    *          The iterator for the collection which contains the needed 
    *          indexes.
    *          
    * @return  List of SQL statements.  Will be empty if no indices are
    *          defined.
    */
   protected List<String> generateSQLDropIndexList(P2JDialect dialect,
                                                   Class<?> dmoIface, 
                                                   Table primary, 
                                                   List<Table> secondaries,
                                                   Iterator<P2JIndex> iter)
   {
      List<String> stmts = new ArrayList<>();
      
      if (!iter.hasNext())
      {
         // If no other indices exist, drop the index on the multiplex ID of the primary table.
         String tableName = primary.getName();
         String idxName = "idx_mpid__" + tableName;
         stmts.add(generateSQLDropIndex(dialect, idxName));

         ((ArrayList<String>) stmts).trimToSize();
      }
      else 
      {
         stmts = super.generateSQLDropIndexList(dialect, dmoIface, primary, secondaries, iter);
      }
      
      return stmts;
   }

   /**
    * This method which will return the computed columns for the required
    * DMO, as retrieved from the dmo_index.xml file.
    * 
    * @param   schema
    *          The schema used for this DMO.
    * @param   dmoIface
    *          The DMO for which to retrieve the computed columns.
    * 
    * @return  A map with the computed columns associated with this DMO.
    */
   protected Map<String, P2JIndexComponent> getComputedCols(String schema, Class<?> dmoIface)
   {
      try
      {
         return DMOIndex.getIndexedCharacterColumns(schema, database, dmoIface);
      }
      catch (PersistenceException exc)
      {
         LOG.error("Error getting computed columns", exc);
         
         return Collections.emptyMap();
      }
   }
   
   /**
    * This method will return the indexes associated with a certain DMO, as
    * retrieved from the dmo_index.xml file.
    * 
    * @param   schema
    *          The schema used for this DMO.
    * @param   dmoIface
    *          The DMO for which to retrieve the indexes.
    * 
    * @return  An iterator for the indexes associated with this DMO.
    */
   protected Iterator<P2JIndex> getIndexes(String schema, Class<?> dmoIface)
   {
      try
      {
         return DMOIndex.databaseIndexes(schema, database, dmoIface);
      }
      catch (PersistenceException exc)
      {
         LOG.error("Error querying index data for " + dmoIface.getName(), exc);
         
         throw new StopConditionException(exc);
      }
   }
}
