/*
** Module   : P2JSQLServer2008Dialect.java
** Abstract : Implementation of P2JDialect for a MS SQL Server 2008 backend
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 VMN 20131012 Created initial version. Adds P2JDialect method implementations to extend
**                  Hibernate's SQLServer2008Dialect.
** 002 OM  20140106 Added support for mixed asc/desc indexes.
** 003 OM  20140410 Fixed support case-sensitive fields in indexes.
**                  Added Override method annotations.
*/

package com.goldencode.p2j.persist.dialect;

import java.sql.Date;
import java.sql.SQLException;
import java.sql.Types;
import java.text.NumberFormat;
import java.util.*;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import com.goldencode.p2j.util.LogHelper;
import org.hibernate.QueryException;
import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.SQLServer2008Dialect;
import org.hibernate.dialect.function.SQLFunction;
import org.hibernate.engine.spi.*;
import org.hibernate.mapping.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.sequence.SequenceHandler;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.util.date;
import com.goldencode.util.StringHelper;

import static com.goldencode.p2j.persist.dialect.DialectHelper.*;

/**
 * Extends Hibernate's <code>SQLServer2008Dialect</code> class to add method implementations
 * required by the {@link P2JDialect} interface.
 * <p>
 * Note:  this implementation expects a directory to be available at the time
 * the class is loaded.
 */
public class P2JSQLServer2008Dialect
extends SQLServer2008Dialect
implements P2JDialect
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(P2JSQLServer2008Dialect.class);

   /** Flag indicating whether to inline a record limit in an SQL SELECT */
   private static final ContextLocal<Boolean> inlineLimit =
           new ContextLocal<Boolean>()
           {
              protected Boolean initialValue() { return Boolean.FALSE; }
           };

   /**
    * Lowest date supported by SQL Server 2008 (January 1, 1753)
    * http://msdn.microsoft.com/en-US/library/ms187819%28v=sql.100%29.aspx
    */
   private static final Date lowDate;

   /**
    * Highest date supported by SQL Server 2008 (December 31, 9999)
    * http://msdn.microsoft.com/en-US/library/ms187819%28v=sql.100%29.aspx
    */
   private static final Date highDate;

   /**
    * Private constant from SQLServer2005Dialect used in sql types mapping with
    * registerColumnType(). We need to overwrite mapping to varchar/varbinary exactly with this
    * constant.
    */
   private static final int MAX_LENGTH = 8000;

   /** Private constant for using in rtrim4gl. */
   private final static String RTRIM_4GL_PART_1 = "substring(";

   /** Private constant for using in rtrim4gl. */
   private final static String RTRIM_4GL_PART_2 = ", 0, len(replace(replace(replace(";

   /** Private constant for using in rtrim4gl. */
   private final static String RTRIM_4GL_PART_3 =
                            ", CHAR(9), ' '), CHAR(10), ' '), CHAR(13), ' ')))";

   /** Private constant for using in extract column. */
   private final static int RTRIM_4GL_PART_1_LENGTH = RTRIM_4GL_PART_1.length();

   /** Private constant for using in extract column. */
   private final static int RTRIM_4GL_PART_2_LENGTH = RTRIM_4GL_PART_2.length();

   /** Private constant for using in extract column. */
   private final static int RTRIM_4GL_PART_3_LENGTH = RTRIM_4GL_PART_3.length();

   static
   {
      // Prepare supported date values.  We use the GMT timezone to define
      // our date range, since this is how com.goldencode.p2j.util.date
      // handles its dates.

      TimeZone tz = TimeZone.getTimeZone("GMT");
      GregorianCalendar cal = new GregorianCalendar(tz);

      cal.set(Calendar.ERA, GregorianCalendar.AD);
      cal.set(Calendar.YEAR, 1753);
      cal.set(Calendar.MONTH, Calendar.JANUARY);
      cal.set(Calendar.DATE, 1);
      lowDate = new Date(cal.getTime().getTime());

      cal.clear();

      cal.set(Calendar.ERA, GregorianCalendar.AD);
      cal.set(Calendar.YEAR, 9999);
      cal.set(Calendar.MONTH, Calendar.DECEMBER);
      cal.set(Calendar.DATE, 31);
      highDate = new Date(cal.getTime().getTime());
   }

   /**
    * Default constructor.  This constructor is for Hibernate's use.  Use code
    * should not construct instances of this class.
    * <p>
    * We provide this constructor to override the superclass' mapping of
    * <code>Types.VARCHAR</code> and <code>Types.VARBINARY</code>.
    */
   public P2JSQLServer2008Dialect()
   {
      super();

      // Override superclass' VARCHAR and VARBINARY mappings, since we always
      // use the unbounded variants of these data types.

      registerColumnType( Types.VARBINARY, MAX_LENGTH, "varbinary" );
      registerColumnType( Types.VARCHAR, MAX_LENGTH, "varchar(8000)" );

      // Maximum SQL Server 2008-2012 NUMERIC column precision is 38.
      registerColumnType(Types.NUMERIC, "numeric(38,$s)");

      registerFunction("rtrim4gl", new Rtrim4GLFunction());
   }

   /**
    * Create a data object, the purpose of which is specific to a particular
    * dialect implementation, to be stored as a context-local object.  This
    * allows a dialect to establish context-local data which is stored by the
    * persistence runtime framework, without the overhead of a separate,
    * context-local variable.  This object will be handed back to the dialect
    * in callback methods when the services of certain hooks are required.
    * <p/>
    * Since the context-local data is specific to a particular dialect
    * implementation, it is optional.  Dialect implementations which do not
    * have a need for this data should simply return <code>null</code>.
    *
    * @param   persistence
    *          Provider of persistence services.
    *
    * @return Dialect-specific, context-local data, or <code>null</code>.
    */
   @Override
   public Object createContext(Persistence persistence)
   {
      return null;
   }
   
   /**
    * A dialect-specific callback hook which is invoked by the persistence
    * runtime framework after a transaction is complete.  The Hibernate
    * session will have been closed by the time this method is invoked.
    *
    * @param   persistence
    *          Provider of persistence services.
    * @param   context
    *          Optional, context-local data created previously by {@link #createContext}.
    * @param   rollback
    *          <code>true</code> if transaction ended with a rollback;
    *          <code>false</code> if transaction ended with a commit.
    *
    * @throws com.goldencode.p2j.persist.PersistenceException
    *          if any persistence-related error occurs.
    */
   @Override
   public void postTransaction(Persistence persistence, Object context, boolean rollback)
   throws PersistenceException
   {
      // do nothing
   }

   /**
    * Retrieve the earliest possible date value which is valid for the
    * backing database.
    *
    * @return Earliest date supported by the backing database.
    */
   @Override
   public Date getLowDate()
   {
      return lowDate;
   }

   /**
    * Retrieve the latest possible date value which is valid for the backing
    * database.
    *
    * @return Latest date supported by the backing database.
    */
   @Override
   public Date getHighDate()
   {
      return highDate;
   }

   /**
    * Format a date object as a string which is appropriate for use in an SQL statement.
    *
    * @param   d
    *          Date to be formatted as a string.
    *
    * @return  Formatted representation of the given date.
    */
   @Override
   public String formatDate(date d)
   {
      NumberFormat nf = NumberFormat.getInstance();

      nf.setMinimumIntegerDigits(4);
      nf.setGroupingUsed(false);

      StringBuilder buf = new StringBuilder();
      int year = d.getYear();

      buf.append(nf.format(Math.abs(year)));
      buf.append('-');

      nf.setMinimumIntegerDigits(2);
      buf.append(nf.format(d.getMonth()));
      buf.append('-');
      buf.append(nf.format(d.getDay()));

      return buf.toString();
   }

   /**
    * Given a column expression as queried from database metadata, extract the
    * root name of a column.  For example, a column name as found within an
    * index in metadata might be represented as:
    * <pre>
    *    upper(rtrim("key"::text, ' \t\n\r'::text))
    * <pre>
    * We are interested only in the <code>key</code> root name, so we strip
    * away the upper() and rtrim() functions, the double quotes, and the
    * <code>::text</code> suffix.
    *
    * @param   expr
    *          Expression containing column name as natively represented by
    *          the database dialect.
    *
    * @return  Root column name.
    */
   @Override
   public String extractColumnName(String expr)
   {
      if (expr.startsWith(INSENSITIVE_CHAR_FIELD))
      {
         expr = expr.substring(INSENSITIVE_CHAR_FIELD.length());
      }
      else if (expr.startsWith(SENSITIVE_CHAR_FIELD))
      {
         expr = expr.substring(SENSITIVE_CHAR_FIELD.length());
      }
      else if (expr.startsWith(CC_PREFIX))
      {
         // has the computed column marker but not case-sensitivity
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Invalid column name: [" + expr + "]." +
                  "Failed to extract the root name of the column.");
         }
      }

      final String upper = "upper(";

      String root = expr.toLowerCase();

      // Remove "upper" function if present (only applicable to text fields).
      if (root.startsWith(upper))
      {
         int len = root.length();
         root = root.substring(upper.length(), len - 1);
      }

      // Remove rtrim4gl function if present (only applicable to text fields).
      if (root.startsWith("substring"))
      {
         root = expr.substring(RTRIM_4GL_PART_1_LENGTH, (expr.length() + RTRIM_4GL_PART_1_LENGTH
                               - RTRIM_4GL_PART_2_LENGTH - RTRIM_4GL_PART_3_LENGTH) / 2);
      }

      // Remove data type suffix.
      int delim = root.indexOf("::");

      root = (delim >= 0 ? root.substring(0, delim) : root);

      // Strip enclosing parentheses.
      while (root.startsWith("(") && root.endsWith(")"))
      {
         root = root.substring(1, root.length() - 1);
      }

      // Strip enclosing double quotes or backticks.
      root = StringHelper.stripEnclosing(root, '"');
      root = StringHelper.stripEnclosing(root, '`');

      return root;
   }

   /**
    * Adjust a character column name appropriate to this dialect, such that it
    * can be inserted into an index.  This may involve wrapping it within one
    * or more functions, prepending a special prefix to designate a computed
    * column, or any other transformation appropriate to the database syntax.
    *
    * @param   name
    *          Base column name.
    * @param   ignoreCase
    *          <code>true</code> to cause case to be ignored, else <code>false</code>.
    *
    * @return  Processed column name.
    */
   @Override
   public String getProcessedCharacterColumnName(String name, boolean ignoreCase)
   {
      // not sure implementation is correct, TODO investigate issue
      StringBuilder buf = new StringBuilder(
            ignoreCase ? INSENSITIVE_CHAR_FIELD : SENSITIVE_CHAR_FIELD);
      
      if (ignoreCase)
      {
         buf.append("upper(");
      }
      addRtrim4gl(buf, name);
      if (ignoreCase)
      {
         buf.append(")");
      }

      return buf.toString();
   }

   /**
    * Report whether the given column name represents an alias or an
    * expression which indicates the column is case insensitive.  Assume the
    * column represented by <code>name</code> is already determined to be the
    * appropriate data type for textual data.
    *
    * @param   name
    *          Column name, alias, or expression.
    *
    * @return  <code>true</code> if <code>name</code> represents case-insensitive data.
    */
   @Override
   public boolean isCaseInsensitiveColumn(String name)
   {
      name = StringHelper.stripEnclosing(name, DOUBLE_QUOTE);
      return name.startsWith(INSENSITIVE_CHAR_FIELD);
   }
   
   /**
    * Report whether the given column name represents an alias or an expression which indicates
    * the column is computed.  Assume the column represented by <code>name</code> is
    * already determined to be the appropriate data type for textual data.
    *
    * @param   name
    *          Column name, alias, or expression.
    *
    * @return  <code>true</code> if <code>name</code> represents a computed column.
    */
   public boolean isComputedColumn(String name)
   {
      // we rely on the fact that both INSENSITIVE_CHAR_FIELD and SENSITIVE_CHAR_FIELD
      // have the same prefix: CC_PREFIX
      return name.startsWith(CC_PREFIX);
   }

   /**
    * Indicate whether this dialect requires computed columns to participate
    * within index definitions, instead of embedded expressions.
    *
    * @return  <code>true</code> to indicate computed columns are required.
    */
   @Override
   public boolean needsComputedColumns()
   {
      return true;
   }

   /**
    * Get the prefix, if any, which should be prepended to computed column and/or property names.
    *
    * @param   caseSensible
    *          <code>true</code> if this is a character field was declared with
    *          <code>CASE-SENSITIVE</code> attribute.
    *
    * @return  <code>null</code> to indicate the backing database does not use computed columns.
    *
    * @see     #needsComputedColumns()
    */
   @Override
   public String getComputedColumnPrefix(boolean caseSensible)
   {
      return caseSensible ? SENSITIVE_CHAR_FIELD : INSENSITIVE_CHAR_FIELD;
   }

   /**
    * Indicate whether this dialect natively supports the overloading of user
    * defined functions by parameter number and data type.
    *
    * @return  false because Server 2008 does not support functions overloading
    *          ("Function names ... must be unique within the database and to its schema.")
    */
   @Override
   public boolean supportsFunctionOverloading()
   {
      return false;
   }

   /**
    * Generate a DDL string which will create a global sequence with the given
    * starting value.
    *
    * @param   name
    *          Sequence name.
    * @param   start
    *          Sequence starting value.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getCreateSequenceString(String name, long start)
   {
      return null;
   }

   /**
    * Generate a DDL string which will create a sequence with given properties.
    *
    * @param   name
    *          Sequence name. Mandatory not null.
    * @param   init
    *          Sequence starting value. Mandatory(not null).
    * @param   increment
    *          Sequence incrementing value. Mandatory(not null).
    * @param   minVal
    *          Sequence minimum value. Optional (may be null).
    * @param   maxVal
    *          Sequence maximum value. Optional (may be null).
    * @param   cycle
    *          Sequence cycling property. Optional, if missing assumed not cycling.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getCreateSequenceString(String name,
                                         Long init,
                                         Long increment,
                                         Long minVal,
                                         Long maxVal,
                                         Boolean cycle)
   {
      return null;
   }

   /**
    * Generate a DDL string which will drop a sequence with given properties.
    *
    * @param   name
    *          The name of the sequence to be dropped.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getDropSequenceString(String name)
   {
      return null;
   }

   /**
    * Generate the appropriate select statement to retrieve the next value of a sequence.
    *
    * @param   sequenceName
    *          The name of the sequence.
    * @param   newVal
    *          The new value to be set.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getSequenceSetValString(String sequenceName, long newVal)
   {
      return null;
   }

   /**
    * Generate the appropriate select statement to reset the value of a sequence.
    *
    * @param   sequenceName
    *          The name of the sequence to be queried.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getSequenceCurrValString(String sequenceName)
   {
      return null;
   }

   /**
    * Returns <code>true</code> if this dialect fully support the sequences as
    * 4GL language (most important: bounded values, cycle). If not fully
    * supported, additional code is needed in P2J sequence class implementation to fix this.
    *
    * @return  <code>false</code> because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public boolean hasFullSequenceSupport()
   {
      return false;
   }

   /**
    * Given an exception, this method will check if it indicates a database
    * connectivity error.
    * <p>
    * For the SQLServer dialect, 08S01 – common state for disconnect and cancel.
    * <p>
    *
    * @param   exc
    *          The exception to be checked.
    *
    * @return  <code>true</code> if the exception indicates a database
    *          connectivity error.
    */
   @Override
   public boolean isConnectionError(Exception exc)
   {
      boolean error = false;

      Throwable next = exc;
      while (next != null && !error)
      {
         if (next instanceof SQLException)
         {
            String state = ((SQLException) next).getSQLState();

            error = "08S01".equals(state);
         }
         next = next.getCause();
      }

      return error;
   }

   /**
    * Indicate whether a query parameter which is part of a range check should
    * be inlined into the where clause, or whether it should be substituted
    * into a prepared statement.  This will result in an "inlined"
    * substitution parameter within a where clause.  For example, given a
    * where clause of:
    * <pre>
    *    where value >= ? and id = ?
    * </pre>
    * along with respective substitution parameters of 45 and 100, the
    * inlined version would read:
    * <pre>
    *    where value >= 45 and id = ?
    * </pre>
    * Only the <code>id</code> substitution parameter of 100 would be
    * substituted into the prepared statement.
    * <p>
    * Inlining in such a way can give some database dialects the opportunity
    * to choose a better query plan, than if the parameter being tested with
    * a range check were to be substituted at statement execution time.
    * <p>
    * Note:  this serves as a hint which may be overridden by runtime logic or
    * other configuration settings.
    *
    * @return  <code>true</code> to inline range check parameters;
    *          <code>false</code> to use normal parameter substitution.
    */
   @Override
   public boolean isQueryRangeParameterInlined()
   {
      return false;
   }

   /**
    * Enable/disable limit clause inlining.  Inlining the limit clause will
    * cause the limit/max value to be hard-coded into the SQL statement,
    * rather than being bound as a substitution parameter.
    * <p/>
    * Dialects are free to treat this method as a hint.
    *
    * @param   on
    *          <code>true</code> to enable inlining;  <code>false</code> to disable it.
    */
   @Override
   public void inlineLimit(boolean on)
   {
      // TODO
   }

   /**
    * Determines whether substitution parameters should be explicitly casted
    * to their datatypes inside "then ? else ?" statement.
    *
    * @return <code>true</code> if they should be casted.
    */
   @Override
   public boolean requiresExplicitCastInsideTernary()
   {
      // TODO
      return false;
   }

   /**
    * Determines whether dialect uses single backslash (e.g. '12\34') into
    * string literals rather than double backslash (e.g. '12\\34').
    *
    * @return <code>true</code> if the dialect uses single backslash.
    */
   @Override
   public boolean usesSingleBackslash()
   {
      // TODO
      return false;
   }

   /**
    * Build the DDL to drop an index based upon the specified definition.
    * The SQL is generated by using the database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be dropped.
    *
    * @return  The DDL to drop the given index.
    */
   @Override
   public String getDropIndexString(P2JIndex index)
   {
      // SQL Server 2008-2012 requires fully qualified index name for drop index.

      return "if exists (select * from sys.indexes where name = '" + index.getName() + "') " +
               "drop index " + index.getTable() + '.' + index.getName();
   }

   /**
    * Build the DDL to create an index based upon the specified definition.
    * The SQL is generated by using a database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be created.
    * @param   unique
    *          If <code>true</code>, create the index as a unique index,
    *          <code>false</code> to omit the unique constraint.
    * @param   toFile
    *          <code>true</code> if the string will be output to a file or
    *          <code>false</code> if it will be directly submitted via JDBC.
    *
    * @return  The DDL which was created.
    */
   @Override
   public String getCreateIndexString(P2JIndex index, boolean unique, boolean toFile)
   {
      // Add primary key column to any index which is not explicitly or implicitly unique.
      // TODO: move this to import ruleset so we don't hardcode primary key column name.
      if (!index.isUnique())
      {
         // the direction of the pk column name is the same with the last component before it.
         // this make this index much easier to fit using the sort clause of the FindQuery, 
         // especially in reverse sorting mode (at runtime I cannot tell is the sorting direction
         // was already reversed during conversion for FIND LAST/FIND PREV structures so I
         // cannot assume the direction of pk id is always ascending; however, if the direction
         // is linked to one of the components of the index, reversing it's sorting direction
         // will also reverse the sort order of the pk id so the query planner will be able to
         // identify the correct index at runtime).
         boolean idDesc = false;
         if (index.size() > 0) // should normally be
         {
            P2JIndexComponent last = index.getComponent(index.size() - 1);
            idDesc = last.isDescending(); 
         }
         index.addComponent(DatabaseManager.PRIMARY_KEY, idDesc, false, false, false);
      }

      // Extract index information into a Hibernate-friendly form.
      String tabName = index.getTable();
      Table table = new Table(tabName);
      List<Column> columns = new ArrayList<>();
      Iterator<P2JIndexComponent> iter = index.components();
      while (iter.hasNext())
      {
         P2JIndexComponent next = iter.next();
         String colName = next.getName();
         if (next.isCharType())
         {
            colName = getComputedColumnPrefix(!next.isIgnoreCase()) + colName;
         }
         
         // to allow asc/desc clauses for the index, we manipulate the column
         // name to add the "desc" direction, if needed; this is safe as the
         // H2 dialect doesn't have quoted columns, so the output DDL will be
         // without quotes
         if (next.isDescending())
         {
            colName = colName + " desc";
         }

         columns.add(new Column(colName));
      }

      // Compose the string.
      String createString =
              Index.buildSqlCreateIndexString(this,
                      index.getName(),
                      table,
                      columns.iterator(),
                      unique,
                      null,
                      null);

      return createString;
   }

   /**
    * Build the DDL to set the formula for the computed columns associated
    * with an indexed, character column.
    * <p>
    * The syntax to set the formula for the computed column is:
    * <code>
    * alter table [table] alter column [column] [type] as [formula]
    * </code>
    *
    * @param   config
    *          The hibernate's configuration object.
    * @param   implClassName
    *          The full name of the implementing class for the DMO.
    * @param   colName
    *          The column name which needs to have a computed version. This is
    *          not the computed column name, but the name of its "parent".
    * @param   ignoreCase
    *          <code>true</code> if this is a character field which has the
    *          "ignore case" flag set.
    *
    * @return  The DDL to alter the computed column, so that its formula is
    *          specified.
    */
   @Override
   public String getAlterComputedColumnString(Configuration config,
                                              String implClassName,
                                              String colName,
                                              boolean ignoreCase)
   {
      final PersistentClass pc = config.getClassMapping(implClassName);
      Table table = pc.getIdentityTable();
      Column computedCol = new Column(getComputedColumnPrefix(!ignoreCase) + colName);

      StringBuilder buf = new StringBuilder();
      if (ignoreCase)
      {
         buf.append("upper(");
      }
      addRtrim4gl(buf, colName);
      if (ignoreCase)
      {
         buf.append(")");
      }

      // the syntax to set the formula for the computed column is:
      // alter table <table> add column <column> as <formula>
      return String.format("alter table %s add %s as %s;",
              table.getQuotedName(this),
              computedCol.getQuotedName(this),
              buf.toString());
   }

   /**
    * Prepare the given metadata query parameter.
    *
    * @param   param
    *          The metadata query parameter.
    *
    * @return the prepared parameter.
    */
   @Override
   public String prepareMetadataParameter(String param)
   {
      // TODO
      return null;
   }

   /**
    * Process the given metadata query result, by making it lowercase.
    *
    * @param   value
    *          The metadata query result.
    *
    * @return  the lowercased result.
    */
   @Override
   public String processMetadataResult(String value)
   {
      // TODO
      return null;
   }

   /**
    * Check if the given value is a descending clause for a column index.
    *
    * @param   value
    *          The value to be checked.
    *
    * @return  <code>true</code> if the value is the "D" or "d" string.
    */
   @Override
   public boolean isMetadataSortDesc(String value)
   {
      // TODO
      return false;
   }

   /**
    * Check if the given value is an ascending clause for a column index.
    *
    * @param   value
    *          The value to be checked.
    *
    * @return  <code>true</code> if the value is the "A" or "a" string.
    */
   @Override
   public boolean isMetadataSortAsc(String value)
   {
      // TODO
      return false;
   }

   /**
    * Get a list of database prepare DDL statements.
    *
    * @param    database
    *           The database which needs to be prepared.
    *
    * @return   an empty list.
    */
   @Override
   public List<String> getDatabasePrepareStatements(Database database)
   {
      return Collections.EMPTY_LIST;
   }

   /**
    * Check if the dialect needs to set an explicit collation during schema
    * DDL generation.
    *
    * @return  always <code>true</code>.
    */
   @Override
   public boolean explicitSetCollation()
   {
      // collation is a property of a database or a column, not of a connection. But we can
      // override the collation on the statement level using the COLLATE statement:
      // SELECT ... FROM ... COLLATE ...
      return false;
   }

   /**
    * Get the statement delimiter for this dialect.
    *
    * @return  the statement delimiter.
    */
   @Override
   public String getDelimiter()
   {
      return ";";
   }

   /**
    * Set any database parameters before, to be available when the database
    * driver is loaded.
    * <p>
    * This implementation does nothing.
    */
   @Override
   public void beforeDriverLoad()
   {
      // no-op
   }

   /**
    * Build the URL to be used for a remote connection, based on the given URL and database.
    *
    * @param   url
    *          The URL as received from the remote server.
    * @param   database
    *          The database to which this URL belongs.
    *
    * @return  The built URL as to be used for a remote connection.
    *
    * @throws  PersistenceException
    *          if the specified connection URL is invalid.
    */
   @Override
   public String buildRemoteURL(String url, Database database)
   throws PersistenceException
   {
      return DialectHelper.buildRemoteURL(url, database);
   }

   /**
    * Creates and returns a <code>SequenceHandler</code> for a SQL Server 2008 database.
    *
    * @param   ldbName
    *          The logical name of the database to create the <code>SequenceHandler</code> for.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It is appeared since SQL Server 2012.
    */
   @Override
   public SequenceHandler createSequenceHandler(String ldbName)
   {
      return null;
   }

   /**
    * Report whether the computed columns should be injected on table creating.
    *
    * @return  <code>false</code> for SQL Server.
    */
   @Override
   public boolean injectComputedColumns()
   {
      return false;
   }

   /**
    * Add sql expression removing trailing ' \t\n\r' symbols from given character expression to
    * given StringBuilder. Method created because sql server rtrim() method removes only
    * spaces.
    *
    * @param   stringBuilder
    *          String builder where resulting sql expression will be added.
    *
    * @param   expr
    *          Character expression.
    */
   private static void addRtrim4gl(StringBuilder stringBuilder, String expr)
   {
      stringBuilder.append(RTRIM_4GL_PART_1);
      stringBuilder.append(expr);
      stringBuilder.append(RTRIM_4GL_PART_2);
      stringBuilder.append(expr);
      stringBuilder.append(RTRIM_4GL_PART_3);
   }

   /**
    * A function which trailing ' \t\n\r' symbols from given character expression.
    */
   private static class Rtrim4GLFunction
   implements SQLFunction
   {
      /**
       * Does this function have any arguments?
       *
       * @return   true because this function requires one parameter.
       */
      public boolean hasArguments()
      {
         return true;
      }

      /**
       * If there are no arguments, are parentheses required?
       *
       * @return True if a no-arg call of this function requires parentheses.
       */
      public boolean hasParenthesesIfNoArguments()
      {
         return true;
      }

      /**
       * The return type of the function.  May be either a concrete type which is preset, or
       * variable depending upon the type of the first function argument.
       * <p/>
       * Note, the 'firstArgumentType' parameter should match the one passed into {@link #render}
       *
       * @param   firstArgumentType
       *          The type of the first argument.
       * @param   mapping
       *          The mapping source.
       *
       * @return  The type to be expected as a return.
       *
       * @throws  org.hibernate.QueryException Indicates an issue resolving the return type.
       */
      public Type getReturnType(Type firstArgumentType, Mapping mapping)
      throws QueryException
      {
         return firstArgumentType;
      }

      /**
       * Render the function call as SQL fragment.
       * <p/>
       * Note, the 'firstArgumentType' parameter should match the one passed into
       * {@link #getReturnType}
       *
       * @param   firstArgumentType
       *          The type of the first argument.
       * @param   arguments
       *          The function arguments.
       * @param   factory
       *          The SessionFactory.
       *
       * @return  The rendered function call
       *
       * @throws  org.hibernate.QueryException Indicates a problem rendering the
       * function call.
       */
      public String render(Type firstArgumentType,
                           List arguments,
                           SessionFactoryImplementor factory)
      throws QueryException
      {
         if ( arguments.size() != 1 )
         {
            throw new QueryException("rtrim4gl() requires one argument");
         }
         StringBuilder buf = new StringBuilder();
         addRtrim4gl(buf, (String) arguments.get(0));

         return buf.toString();
      }
   }
}
